"""This part contains some general support function to deal with data.
such as one function could be used in both preprocessing and transforms
Make sure the original source in function. (the files are not import from other code in this package.)
"""
